{
  Copyright 2011-2012 Joe Justice (joe@joejustice.org)
 
  This file is part of LightNet: Cloud.

  LightNet: Cloud is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  LightNet: Cloud is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with LightNet: Cloud, see files COPYING.txt and COPYING.LESSER.txt 
  If not, see <http://www.gnu.org/licenses/>.

----------------------------------------------------------------------------

The file manager. Handles all file operations for LnC. All FTP operations
downloading, uploading and making directories along with deleting local files.

}

unit lnc_fm_unit;

{$mode objfpc}{$H+}

interface
uses
  Classes, SysUtils, FileUtil, lFTP, lNet, lNetComponents;
  
type
  TLnC_FMJob=(
    fmNone,
    fmGet, // Initates the the get
    fmGetting, // While getting
    fmGot, // FTP has finished sending
    fmPut,
    fmPutDone,
    fmMkDir,
    fmList,
    fmListing,
    fmFTPDel,
    fmDel
  );

  TLnC_FM = class
   private
    FCon: TLFTPClientComponent;  // the FTP connection itself
    RecFile: TFileStream; // file stream for local files into
    FatalError: Boolean;     // used as controller of the main loop
    Host, Name, Pass: string; Port: word; // FTP server info
    Jobs: TStringList; // List of jobs sent externally to be parsed and executed here
    LnC_FMJob: TLnC_FMJob; // Current job type
    FMJobLine1, FMJobLine2: string; // Jobs line info
    RCurrentBytes, RTotalBytes: integer;  // Current and Total bytes for receive file
    SCurrentBytes, STotalBytes: integer;  // Current and Total bytes for send file
    ftpConnected :boolean;
    RetryCounter, DisConCounter: integer;
    CurrentDataMessage: string;
    procedure FMJobs_Trigger;
    procedure FMJobs_Process;
    procedure FMJobs_Finish;
    procedure fmPut_Process;
    procedure fmGet_Process;
    procedure fmList_Process;
    function ftpExtractFileName(const s: string): string;
    function ftpExtractFilePath(const s: string): string;
    procedure ErrorSet(const s1, s2: string);
    // FTP Events
    procedure OnConnect(aSocket: TLSocket);
    procedure OnReceive(aSocket: TLSocket);
    procedure OnControl(aSocket: TLSocket);
    procedure OnSent(aSocket: TLSocket; const Bytes: Integer);
    procedure OnError(const msg: string; aSocket: TLSocket);
    procedure OnSuccess(aSocket: TLSocket; const aStatus: TLFTPStatus);
    procedure OnFailure(aSocket: TLSocket; const aStatus: TLFTPStatus);
   public
    Listing, // File Listing
    ListingTest, // File Listing variable used to check and see if the Listing has changed
    Messages: string; //Messages to be used in UI
    Progress: integer;  //Value for progress bar
    MessageLine1, MessageLine2: string; // UI strings for feedback
    LoginSuccess: boolean;
    property Connected: Boolean read ftpConnected;
    constructor Create;
    destructor Destroy; override;
    procedure Run(const ftpHost, ftpPort, ftpName, ftpPass: string); // this is where the main loop is
    procedure ErrorReset;
    function HasFatalError: boolean; //FTP has a fatal error
    function IsBusy: boolean; //FTP is doing something
    function IsGetting: boolean; //FTP is downloading
    function IsPutting: boolean; //FTP is uploading
    procedure FMJobs_Add (const s: string);
    function HasJobsPending: boolean;
  end;

implementation

const
  FM_PartExt = '.part';
  MaxRetry = 20;
  DisConSec = 600;

constructor TLnC_FM.Create;
begin
  FCon := TLFTPClientComponent.Create(nil);
  FCon.Timeout := 50; // 50 milliseconds is nice to save CPU but fast enough to be responsive to humans
  FCon.OnConnect := @OnConnect; // assign all events
  FCon.OnReceive := @OnReceive;
  FCon.OnControl := @OnControl;
  FCon.OnSent := @OnSent;
  FCon.OnError := @OnError;
  FCon.OnSuccess := @OnSuccess;
  FCon.OnFailure := @OnFailure;
  Listing := '';
  ListingTest := '*';
  Messages := '';
  MessageLine1 := 'Ready';
  MessageLine2 := '';
  Progress := 0;
  Jobs := TStringList.Create;
  Jobs.Sorted := false;
  LnC_FMJob := fmNone;
  FMJobLine1 := '';
  FMJobLine2 := '';
  FatalError := False;
  LoginSuccess := False;
  Host := '';
  ftpConnected := False;
  RetryCounter := 0;
  DisConCounter := 0;
  CurrentDataMessage := '';
end;

destructor TLnC_FM.Destroy;
begin
  FCon.Free;
end;

procedure TLnC_FM.FMJobs_Trigger;
begin
  DisConCounter := 0;
  if ((LnC_FMJob = fmFTPDel) or (LnC_FMJob = fmDel)) and (Jobs.Count = 0) then
    begin
      LnC_FMJob := fmNone; // Deleting is finished
      MessageLine1 := 'Finished';
      MessageLine2 := '';
      Progress := 0;
    end;

  if LnC_FMJob = fmListing then //Pass one cycle to complete any updating
     LnC_FMJob := fmNone
  else
    begin
      if (not HasFatalError) and (Jobs.Count <> 0) then
      begin
        FMJobs_Process;
        if (FMJobLine2 <> '') and (LnC_FMJob <> fmNone) then
          begin
            case LnC_FMJob of
              fmGet:
                begin
                  if CurrentDataMessage = '' then
                    begin
                      CurrentDataMessage := '*'; //Listing sets data message from *, if it has changed then the data message has been received.
                      FCon.List(FMJobLine2);
                    end;
                  fmGet_Process; // Download a file
                end;
              fmList: if ftpConnected then fmList_Process; // List a directory
              fmPut:
              begin
                // Start upload a process
                FCon.ChangeDirectory(ftpExtractFilePath(FMJobLine2));
              end;
              fmMkDir: // Make a directory
              begin
                DisConCounter := 0;
                FCon.MakeDirectory(FMJobLine2);
                MessageLine1 := 'Creating Directory:';
                MessageLine2 := FMJobLine2;
                Progress := 100;
                FMJobs_Finish;
                LnC_FMJob := fmNone;
                FMJobs_Trigger;
              end;
              fmFTPDel: // Make a directory
              begin
                DisConCounter := 0;
                if FMJobLine1 = 'file' then
                  FCon.DeleteFile(FMJobLine2);
                if FMJobLine1 = 'dir' then
                  FCon.RemoveDirectory(FMJobLine2);
                MessageLine1 := 'Deleteing:';
                MessageLine2 := FMJobLine2;
                Progress := 100;
                FMJobs_Finish;
              end;
            end;
          end;
      end;
    end;
end;

procedure TLnC_FM.FMJobs_Process;

var
  JobDelim: TStringList;

begin
  LnC_FMJob := fmNone; //Assume no job to start
  if (Jobs.Count <> 0) then
    begin
      JobDelim := TStringList.Create;
      JobDelim.Delimiter := '|';
      JobDelim.StrictDelimiter := True;
      JobDelim.DelimitedText := Jobs[0];

      if JobDelim[0] = 'get' then LnC_FMJob := fmGet;
      if JobDelim[0] = 'list' then LnC_FMJob := fmList;
      if JobDelim[0] = 'put' then LnC_FMJob := fmPut;
      if JobDelim[0] = 'mkdir' then LnC_FMJob := fmMkDir;
      if JobDelim[0] = 'ftpdel' then LnC_FMJob := fmFTPDel;

      FMJobLine1 := JobDelim[1];
      FMJobLine2 := JobDelim[2];

      if JobDelim[0] = 'del' then
        begin
          LnC_FMJob := fmDel;
          DeleteFile(JobDelim[1]);
          FMJobs_Finish;
          Messages := Messages + 'LnC: Deleted: ' + ExtractFileName(JobDelim[1]) + LineEnding;
          MessageLine1 := 'Finished';
          MessageLine2 := '';
          Progress := 0;
        end;

      JobDelim.Free;
    end;
end;

procedure TLnC_FM.FMJobs_Finish;
begin
//  Messages := Messages + 'DONE! ' + Jobs.ValueFromIndex[0] + LineEnding;
  Jobs.Delete(0);
  RetryCounter := 0;
end;

procedure TLnC_FM.FMJobs_Add (const s: string);
var
  i:integer;

begin
  if Jobs.IndexOf(s) < 0 then Jobs.Add(s);

{//Feedback
  if Jobs.Count > 0 then
    begin
  Messages := Messages + '--------------------' + LineEnding;
  for i := 0 to (Jobs.Count - 1) do
    Messages := Messages + Jobs.ValueFromIndex[i] + LineEnding;
  Messages := Messages + '--------------------' + LineEnding;
    end;}
end;

procedure TLnC_FM.fmList_Process;
begin
  Listing := '';
  ListingTest := '*'; // Set to something other than '' for testing
  FCon.List(FMJobLine2);
  MessageLine1 := 'Retrieving Directory Listing';
  MessageLine2 := FMJobLine2;
  Progress := 100;
  FMJobs_Finish;
end;

procedure TLnC_FM.fmGet_Process;

var
  FTPList: TStringList;

begin
  Inc (RetryCounter);
  FTPList := TStringList.Create;
//Messages := Messages + CurrentDataMessage;
  //Listing sets data message from *, if it has changed then the data message has been received.
  if CurrentDataMessage <> '*' then
    begin
      FTPList.DelimitedText := CurrentDataMessage;
      if FTPList.Count > 4 then RTotalBytes := StrToInt(FTPList[4]);
      RCurrentBytes := 0;
      CurrentDataMessage := ''; //Reset data message check for next download
      //Wipe out local files if they already exist
      if FileExistsUTF8(FMJobLine1 + FM_PartExt) then DeleteFileUTF8(FMJobLine1 + FM_PartExt);
      if FileExistsUTF8(FMJobLine1) then DeleteFileUTF8(FMJobLine1);
    try
      //Start file stream, add part extention
      RecFile := TFileStream.Create(FMJobLine1 + FM_PartExt, fmCreate or fmOpenWrite);
      //Start FTP retrieve of file
      FCon.Retrieve(FMJobLine2);
      LnC_FMJob := fmGetting;
      Messages := Messages + 'LnC: Downloading: ' + ftpExtractFileName(FMJobLine2) + LineEnding;
    except
      on E : Exception do
        begin
          Messages := Messages + 'LnC: Problem downloading file ' + FMJobLine2 + '. Will retry.' + LineEnding;
          LnC_FMJob := fmNone;
          //Rety logic and error setting
          if RetryCounter > MaxRetry then
            ErrorSet('FTP ERROR!', 'Can not download ' + ftpExtractFileName(FMJobLine2));
        end;
    end;
    end
  else //Data message was either blank or has not been recived yet
    begin
      //List was did not have the file that is suppose to be downloaded in it
      LnC_FMJob := fmNone;
      Messages := Messages + 'LnC: Waiting on file data on ' + ftpExtractFileName(FMJobLine2) + LineEnding;
      //If this is past max rety then assume file is not there and set error
      if RetryCounter > MaxRetry then
        ErrorSet('FTP ERROR!', 'File ' + ftpExtractFileName(FMJobLine2) + ' not found');
    end;
  FTPList.Free;
end;

procedure TLnC_FM.fmPut_Process;

var
  SizeFile: TFileStream;

begin
  Inc (RetryCounter);
  if FileExistsUTF8(FMJobLine1) then
    begin
      try
        SizeFile := TFileStream.Create(FMJobLine1, fmOpenRead);
        STotalBytes := SizeFile.Size;
        FreeAndNil(SizeFile);
      except
        on E : Exception do
          begin
            STotalBytes := -1;
            Messages := Messages + 'LnC: Problem uploading file ' + FMJobLine2 + '. Will retry.' + LineEnding;
            LnC_FMJob := fmNone;
            //Rety logic and error setting
            if RetryCounter > MaxRetry then
              ErrorSet('FTP ERROR!', 'Can not upload ' + ftpExtractFileName(FMJobLine2));
          end;
      end;
      //Reading the file did not bring up an execption; okay to upload
      if STotalBytes >= 0 then
        begin
          SCurrentBytes := 16384;
          FCon.Put(FMJobLine1);
          Messages := Messages + 'LnC: Uploading: ' + ExtractFileName(FMJobLine1) + LineEnding;
        end;
    end
    else
      begin
        Messages := Messages + 'LnC: File not found ' + FMJobLine2 + LineEnding;
        LnC_FMJob := fmNone;
        FMJobs_Finish;
        FMJobs_Trigger;
      end;
end;

procedure TLnC_FM.OnConnect(aSocket: TLSocket);
begin
  FCon.Authenticate(Name, Pass);
  FCon.Binary := True; // set binary mode, others are useless anyhow
end;

function TLnC_FM.ftpExtractFileName(const s: string): string;
var
  i:integer;
begin
  i := Length(s);
  repeat
    i := i - 1;
  until (i < 0) or (s[i] = '/');
  if i > 0 then Result := Copy(s, i + 1, Length(s))
  else Result := s;

end;

function TLnC_FM.ftpExtractFilePath(const s: string): string;
var
  i:integer;
begin
  i := Length(s);
  repeat
    i := i - 1;
  until (i < 0) or (s[i] = '/');
  if i > 0 then
    Result := Copy(s, 1, i)
  else
    Result := '/';
end;

procedure TLnC_FM.OnReceive(aSocket: TLSocket);

var
  i: Integer;
  Buf: array[0..65535] of Byte;
  s, CurrentBytesStr, TotalBytesStr: string;

begin
  DisConCounter := 0;
  if (LnC_FMJob = fmGetting) or (LnC_FMJob = fmGot) then
    begin // getting file, save to file
      i := FCon.GetData(Buf, 65535);
      inc (RCurrentBytes, i);
      if i > 0 then
        begin
          RecFile.Write(Buf, i);
          if RTotalBytes > 0 then Progress := Round(RCurrentBytes / RTotalBytes * 100);
          MessageLine1 := 'Downloading ' + FMJobLine2;

          CurrentBytesStr := FloatToStrF(RCurrentBytes / 1000000, ffNumber, 0, 1) + ' MB';
          TotalBytesStr := FloatToStrF(RTotalBytes / 1000000, ffNumber, 0, 1) + ' MB';
          if RTotalBytes > 0 then
            MessageLine2 := 'Downloading ' + IntToStr(Round(100 * (RCurrentBytes / RTotalBytes))) +
                  '% | ' + CurrentBytesStr + ' of ' + TotalBytesStr;
        end;
      if RCurrentBytes >= RTotalBytes then
        begin
          //Finish up
          FreeAndNil(RecFile);
          //Move downloaded file to correct name
          if FileExistsUTF8(FMJobLine1) then DeleteFileUTF8(FMJobLine1);
          RenameFileUTF8(FMJobLine1 + FM_PartExt, FMJobLine1);
          LnC_FMJob := fmNone;
          FMJobs_Finish;
          FMJobs_Trigger;
        end;
      end
    else
      begin
        s := FCon.GetDataMessage;
        if Length(s) > 0 then
          begin
            CurrentDataMessage := s;
            if (LnC_FMJob = fmList) or (LnC_FMJob = fmListing) then
              begin
                Listing := Listing + CurrentDataMessage;
                CurrentDataMessage := ''; //Reset data message check for next download
              end
            else
              begin
                Messages := Messages + CurrentDataMessage;
              end;
          end;
      end;
end;

procedure TLnC_FM.OnSent(aSocket: TLSocket; const Bytes: Integer);

var
  CurrentBytesStr, TotalBytesStr: string;

begin
  DisConCounter := 0;
  if Bytes > 0 then
    begin
      Inc(SCurrentBytes, Bytes);
      if STotalBytes > 0 then Progress := Round(SCurrentBytes / STotalBytes * 100);
      MessageLine1 := 'Uploading ' + ExtractFileName(FMJobLine1);

      CurrentBytesStr := FloatToStrF(SCurrentBytes / 1000000, ffNumber, 0, 1) + ' MB';
      TotalBytesStr := FloatToStrF(STotalBytes / 1000000, ffNumber, 0, 1) + ' MB';
      if STotalBytes > 0 then
        MessageLine2 := 'Uploading ' + IntToStr(Round(100 * (SCurrentBytes / STotalBytes))) +
          '% | ' + CurrentBytesStr + ' of ' + TotalBytesStr;
    end;

    if (STotalBytes <= 16384) or (SCurrentBytes >= STotalBytes) then
      begin
        MessageLine1 := 'Finished upload of ' + ExtractFileName(FMJobLine1);
        MessageLine2 := '';
        Progress := 100;
        LnC_FMJob := fmPutDone;
      end;
end;

procedure TLnC_FM.OnControl(aSocket: TLSocket);
var
  s: string;
  MsgParse: TStringList;
  i: integer;
  Done226: Boolean;

begin
  Done226 := false;
  if FCon.GetMessage(s) > 0 then // if we got some new message about FTP status, write it
    begin
      MsgParse := TStringList.Create;
      MsgParse.Text := s;
      if MsgParse.Count > 0 then
        for i := 0 to MsgParse.Count - 1 do
          begin
            if Length (MsgParse.ValueFromIndex[i]) > 0 then
              Messages := Messages +  MsgParse.ValueFromIndex[i] + LineEnding;
            //Username or password success
            if Copy(MsgParse.ValueFromIndex[i], 1, 3) = '230' then
              begin
                LoginSuccess := True;
              end;
            // Delete was successful
            if (Copy(MsgParse.ValueFromIndex[i], 1, 3) = '250') and (LnC_FMJob = fmFTPDel) then
              begin
                FMJobs_Trigger;
              end;
            // Directory change successful, Put a file
            if (Copy(MsgParse.ValueFromIndex[i], 1, 3) = '250') and (LnC_FMJob = fmPut) then
              begin
                fmPut_Process;
              end;
            //Set error of Username or password are wrong
            if Copy(MsgParse.ValueFromIndex[i], 1, 3) = '530' then
              begin
                ErrorSet('FTP ERROR!', 'Username or password incorrect');
                LoginSuccess := False;
              end;
            if (Copy(MsgParse.ValueFromIndex[i], 1, 3) = '226') and (not Done226) then
              begin
                if LnC_FMJob = fmList then LnC_FMJob := fmListing;
                // FTP has finished sending file
                if (LnC_FMJob = fmGetting) then
                  begin
                    LnC_FMJob := fmGot;
                    Done226 := True;
                  end;
                if (LnC_FMJob = fmPutDone) then
                  begin
                    LnC_FMJob := fmNone;
                    FMJobs_Finish;
                    FMJobs_Trigger;
                    Done226 := True;
                  end;
                if (LnC_FMJob = fmNone) then
                  begin
                    LnC_FMJob := fmNone;
                    FMJobs_Trigger;
                    Done226 := True;
                  end;
              end;
        end;
      MsgParse.Free;
    end;
end;

procedure TLnC_FM.OnError(const msg: string; aSocket: TLSocket);
begin
// Need more complete error checking. Right now assumes all are fatal unless told to ignore
//  Messages := Messages + msg + LineEnding;
  if CurrentDataMessage <> '*' then // Needed for FileZilla which returns error with empty listing
    ErrorSet('FTP ERROR!', 'Check settings and network connection. Error.');
end;

procedure TLnC_FM.OnFailure(aSocket: TLSocket; const aStatus: TLFTPStatus);

var
  s:string;

begin
  if CurrentDataMessage <> '*' then // Needed for FileZilla which returns error with empty listing
    ErrorSet('FTP FAILURE!', 'Check settings and network connection. Failure.');
end;

procedure TLnC_FM.OnSuccess(aSocket: TLSocket; const aStatus: TLFTPStatus);

begin
  case aStatus of
    fsList:;
    fsType:
      begin
//      Messages := Messages + 'Ready for FTP Command' + LineEnding;
      end;
// Other possible statuses to use for debugging
    fsRetr:; //Messages := Messages + 'fsRetr'+ LineEnding;
    fsStor:; //Messages := Messages + 'fsStor'+ LineEnding;
    fsNone:; //Messages := Messages + 'fsNone' + LineEnding;
    fsCon:; //Messages := Messages + 'fsCon'+ LineEnding;
    fsUser:; //Messages := Messages + 'fsUser'+ LineEnding;
    fsPass:; //Messages := Messages + 'fsPass'+ LineEnding;
    fsPasv:; //Messages := Messages + 'fsPasv'+ LineEnding;
    fsPort:; //Messages := Messages + 'fsPort'+ LineEnding;
    fsCWD:; //Messages := Messages + 'fsCWD'+ LineEnding;
    fsMKD:; //Messages := Messages + 'fsMKD'+ LineEnding;
    fsRMD:; //Messages := Messages + 'fsRMD'+ LineEnding;
    fsDEL:; //Messages := Messages + 'fsDEL'+ LineEnding;
    fsRNFR:; //Messages := Messages + 'fsRNFR'+ LineEnding;
    fsRNTO:; //Messages := Messages + 'fsRNTO'+ LineEnding;
    fsSYS:; //Messages := Messages + 'fsSYS'+ LineEnding;
    fsFeat:; //Messages := Messages + 'fsFeat'+ LineEnding;
    fsPWD:; //Messages := Messages + 'fsPWD'+ LineEnding;
    fsHelp:; //Messages := Messages + 'fsHelp'+ LineEnding;
    fsLast:; //Messages := Messages + 'fsLast'+ LineEnding;
  end;
end;

function TLnC_FM.IsBusy: boolean; //FTP is doing something

begin
  Result := False;
  if LnC_FMJob <> fmNone then Result := True;
  if HasFatalError then Result := True;
end;

function TLnC_FM.IsGetting: boolean; //FTP is downloading
begin
  if FCon.CurrentStatus = fsRetr  then
    Result := True
  else
    Result := False;
end;

function TLnC_FM.IsPutting: boolean; //FTP is uploading
begin
  if FCon.CurrentStatus = fsStor then
    Result := True
  else
    Result := False;
end;

procedure TLnC_FM.ErrorSet(const s1, s2: string);
begin
  FatalError := True;
  Jobs.Clear;
  RetryCounter := 0;
  LnC_FMJob := fmNone;
  if (Assigned(RecFile)) then FreeAndNil(RecFile);
  MessageLine1 := s1;
  MessageLine2 := s2;
  ftpConnected := False;
  Progress := 0;
  DisConCounter := 0;
  CurrentDataMessage := ''; //Reset data message check for next download
  Listing := '';
  ListingTest := '*';
  Host := ''; //Reset host to force a reconnect
end;

procedure TLnC_FM.ErrorReset;

begin
  if HasFatalError then
    begin
      Jobs.Clear;
      FatalError := False;
      ftpConnected := False;
      MessageLine1 := 'Ready';
      MessageLine2 := '';
    end;
end;
function TLnC_FM.HasJobsPending: boolean;
begin
  if Jobs.Count > 0 then
    Result := true
  else
    Result := false;
end;

function TLnC_FM.HasFatalError: boolean; //FTP has a fatal error
begin
  if FatalError then
    Result := True
  else
    Result := False;
end;

procedure TLnC_FM.Run(const ftpHost, ftpPort, ftpName, ftpPass: string);

begin
  // Last command was a local deletion
  if LnC_FMJob = fmDel then
    begin
        FMJobs_Trigger; // All finished
    end;
  // Determine if file listing is still being recieved
  if LnC_FMJob = fmListing then
    if Listing <> ListingTest then
      begin // More data has been recived in the file listing
        ListingTest := Listing; // Reset for next test
      end
    else
      begin // The file listing has not changed
        if Listing = '' then Listing := 'empty';
        FMJobs_Trigger; // All finished
      end;

  if ftpConnected then Inc(DisConCounter);
  if DisConCounter > DisConSec then
    begin
      ErrorSet('Disconnected', 'Lack of activity');
    end;

  if not IsBusy then
    begin
      if (Host <> ftpHost) or
         (Port <> Word(StrToInt(ftpPort))) or
         (Name <> ftpName) or
         (Pass <> ftpPass) then
        begin
          //FCon.Disconnect(False);
          ErrorReset;
          LoginSuccess := False;
          ftpConnected := False;
        end;
      Host := ftpHost;
      Port := Word(StrToInt(ftpPort));
      Name := ftpName;
      Pass := ftpPass;

      if not FatalError then
        begin
          if Jobs.Count > 0 then
            begin
              if LoginSuccess then
                begin
                  FMJobs_Trigger;
                end
              else  //Trigger connection
                if not ftpConnected then
                  begin
                    FCon.Connect(Host, Port);
                    MessageLine1 := 'Connecting to FTP Server';
                    MessageLine2 := '';
                    ftpConnected := true;
                  end;
            end
          else
            if ftpConnected then
              begin
                MessageLine1 := 'Connected to FTP Server';
                MessageLine2 := '';
                Progress := 0;
              end;
        end;
    end;
end;

end.

