{
  Copyright 2011-2012 Joe Justice (joe@joejustice.org)
 
  This file is part of LightNet: Cloud.

  LightNet: Cloud is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  LightNet: Cloud is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with LightNet: Cloud, see files COPYING.txt and COPYING.LESSER.txt 
  If not, see <http://www.gnu.org/licenses/>.

----------------------------------------------------------------------------

Performs all operations on the scene files and content archives, including 
figuring out what to download, telling the file manger to download it, 
manipulating it and telling the file manger to upload the output.

}

unit lwscenes_unit;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, IniFiles, FileUtil, lnczipper, lnc_utils;

type
  TLWScene = class(TObject)
  public
    //Message for log
    LogMsg: string;

    //Data from the server
    SceneID, SceneContentDir,SceneName, BaseSceneName, StartFrame, EndFrame, ContentDateTimeStamp: string;
    LnCCommandFromServer: string;

    //Communications with LnC
    function IsOperational: boolean;
    function IsTryingToGetSceneFiles: boolean;
    function HasGatheredScene: boolean;
    function ReadyToGather: boolean;
    function IsFinished: boolean;

    procedure StatusUpdate;
    function GatherFrames(const SentFrameCount: integer): integer;
    //Returns number of frames gathered

    procedure Finish;
    procedure ResetScene;
    procedure Uploadlog(logfile: string);

    constructor Create;
    destructor Destroy; override;

  private
    Status:  integer;
    { Status values:
      Using units of 10 so as to have room for various sub-statuses
      0 - Not Operational
      1 - FTP error
      2 - Localization error
      10 - Ready to Gather frames
      20 - Localizing scene and it's output
      21 - Localizing content dir
      30 - Waiting on render
      40 - Send uploading message to log
      41 - Uploading frames
      42 - Wait for next command after upload
      41 - Finished
      50 - Reset data
    }
    SceneDled: Boolean;
    ftpSubDir, LocalWorkingDir: string;
    ftpSceneName, ftpSceneContentDirZip, SceneContentDirZip:string;
  end;

implementation

uses lightnet_cloud_form;

constructor TLWScene.Create;
begin
  Status := 0;
  LogMsg := '';
  SceneID    := '';
  SceneContentDir := '';
  SceneName  := '';
  BaseSceneName := '';
  StartFrame := '';
  EndFrame   := '';
  ftpSceneName := '';
  ftpSceneContentDirZip := '';
  SceneContentDirZip := '';
  SceneDled := False;
  ContentDateTimeStamp := '';
end;

destructor TLWScene.Destroy;
begin
end;

procedure TLWScene.StatusUpdate;

  function StripDate(const fn:string): string;
  var
    s1, s2: string;
    i1, i2: integer;
  begin
    i1 := Length(fn);
    while (fn[i1] <> ')') or (i1 = 1) do i1 := i1 - 1;
    i1 := i1 + 1; //Move on more character forward

    i2 := i1 - 1;
    while (fn[i2] <> '(') or (i2 = 1) do i2 := i2 - 1;
    i2 := i2 - 1; //Move on more character back

    if i2 > 0 then
      begin
        s1 := Copy(fn, 1, i2);
        if i1 < Length(fn) then s2 := Copy(fn, i1, Length(fn)) else s2 := '';
        Result := s1 + s2;
      end
    else Result := fn;
  end;

var
  UnZipper: TUnZipper;
  Outputdir, OutputdirRGB, OutputdirAlpha:string;
  ftpOutputdir, ftpOutputdirRGB, ftpOutputdirAlpha:string;
  LWSceneFile: TStringList;
  ContetGetTime: TextFile;
  strim, i, i2, IOR: integer;
  s, s1, s2, OutPrefixRGB, OutPrefixAlpha: string;
  OutputFiles: TSearchRec;
  NoFilesLeft: boolean;

begin
  Outputdir := '';
  OutputdirRGB := '';
  OutputdirAlpha := '';
  i := 0;

  //Check to be sure Status is > 9 before doing anything else
  if Status > 9 then
  begin
    case Status of
    10: //Does nothing but wait for the GatherFrame call from LnC
      begin
        //Do Nothing
      end;
    20:
      begin //Localize scene data
        if FileExistsUTF8(SceneName) then
          begin
            //Set output in scene file to local dir
            Outputdir := LnC_Settings.LocalWorkingDir + PathDelim + LnC_GetOutputDir(BaseSceneName, 'root');
            OutputdirRGB := LnC_Settings.LocalWorkingDir + PathDelim + LnC_GetOutputDir(BaseSceneName, 'rgb');
            OutputdirAlpha := LnC_Settings.LocalWorkingDir + PathDelim + LnC_GetOutputDir(BaseSceneName, 'alpha');
            //Create the output dirs if they don't already exist
            if not DirectoryExistsUTF8(OutputDir) then
              begin
                CreateDir (OutputDir);
                CreateDir (OutputDirRGB);
                CreateDir (OutputDirAlpha);
              end;
          //Change the output directory in the Lightwave Scene file to the local directory
            LWSceneFile := TStringList.Create;
            LWSceneFile.LoadFromFile(SceneName);
          //Knock off the .lws from the scene name
            strim := Length(BaseSceneName);
            strim := strim - 4;
            s := Copy(BaseSceneName, 1, strim);
            s := StripDate(s);
            //Set the RGB and Alpha Prefix names
            OutPrefixRGB := s + '_rgb_';
            OutPrefixAlpha := s + '_alpha_';

            {If the lnc_preserveOutput variable was set to 2 by the LScript, it means the user
            wanted to preserve the output prefix. So here we change OutPrefixRGB and OutPrefixAlpha.
            If none of the conditions are right, then the LnC default output prefixes will just get
            passed along unchanged }
            if LwSceneFile.IndexOf('lnc_preserveOutput 2') > 0 then
              begin
                if LwSceneFile.IndexOf('SaveRGB 1') > 0 then
                  begin
                    i := LwSceneFile.IndexOf('SaveRGB 1') + 1;
                    s := Copy (LwSceneFile[i], 20, Length(LwSceneFile[i]));
                    OutPrefixRGB := ExtractFileName(s);
                  end;
                if LwSceneFile.IndexOf('SaveAlpha 1') > 0 then
                  begin
                    i := LwSceneFile.IndexOf('SaveAlpha 1') + 1;
                    s := Copy (LwSceneFile[i], 22, Length(LwSceneFile[i]));
                    OutPrefixAlpha := ExtractFileName(s);
                  end;
              end;
            //Insert Output RGB Prefix
            if LwSceneFile.IndexOf('SaveRGB 1') > 0 then
              begin
                i := LwSceneFile.IndexOf('SaveRGB 1');
                i:= i + 1;
                LWSceneFile.Delete(i);
                LWSceneFile.Insert(i, 'SaveRGBImagesPrefix ' + OutputDirRGB + PathDelim + OutPrefixRGB);
              end;
            //Insert Output Alpha Prefix
            if LwSceneFile.IndexOf('SaveAlpha 1') > 0 then
              begin
                i := LwSceneFile.IndexOf('SaveAlpha 1');
                i:= i + 1;
                LWSceneFile.Delete(i);
                LWSceneFile.Insert(i, 'SaveAlphaImagesPrefix '+ OutputDirAlpha + PathDelim + OutPrefixAlpha);
              end;

            LWSceneFile.SaveToFile(SceneName);
            LWSceneFile.Free;

            //Set the status to ready to gather
            Status := 10;
            //Scene has now been downloaded set for proper render path
            SceneDled := True;
          end
        else
          begin
            if not SceneDled then LnC_FM.FMJobs_Add('get|' + SceneName + '|' + ftpSceneName);
          end;
      end;
    21:
      begin //Localize content
        if FileExistsUTF8(SceneContentDirZip) then
          begin
            //Set date and time stamp for zip content dir
            DateSeparator := '-';
            TimeSeparator := '_';
            ContentDateTimeStamp := '-' + DateTimeToStr(Now);
            DateSeparator := '/'; //Reset DateSeparator variable
            TimeSeparator := ':'; //Reset TimeSeperator variable

            Status := 10;
            UnZipper := TUnZipper.Create;
            try
              UnZipper.FileName := SceneContentDirZip;
              UnZipper.OutputPath := SceneContentDir + ContentDateTimeStamp;
              UnZipper.Examine;
              UnZipper.UnZipAllFiles;
            finally
              UnZipper.Free;
            end;
            if not DirectoryExistsUTF8(SceneContentDir + ContentDateTimeStamp) then
              begin
                Status := 2;
                LogMsg := 'ERROR: There was a problem localizing the content';
              end;
            if (DirectoryExistsUTF8(LnC_Settings.LocalCache)) and
               (not FileExistsUTF8(LnC_Settings.LocalCache + PathDelim + BaseSceneName + PathDelim + ExtractFilename(SceneContentDirZip))) then
              begin
                if not DirectoryExistsUTF8(LnC_Settings.LocalCache + PathDelim + BaseSceneName) then CreateDirUTF8(LnC_Settings.LocalCache + PathDelim + BaseSceneName);
                CopyFile (SceneContentDirZip, LnC_Settings.LocalCache + PathDelim + BaseSceneName + PathDelim + ExtractFilename(SceneContentDirZip) + '.part');
                RenameFileUTF8(LnC_Settings.LocalCache + PathDelim + BaseSceneName + PathDelim + ExtractFilename(SceneContentDirZip) + '.part', LnC_Settings.LocalCache + PathDelim + BaseSceneName + PathDelim + ExtractFilename(SceneContentDirZip));
                //Get rid of the info file
                if FileExistsUTF8(LnC_Settings.LocalCache + PathDelim + BaseSceneName + PathDelim + 'download.info') then
                  DeleteFileUTF8(LnC_Settings.LocalCache + PathDelim + BaseSceneName + PathDelim + 'download.info');
              end;
          end
        else
          begin
            //If the zip file is in the content cache dir, then simply copy it over
            if FileExistsUTF8(LnC_Settings.LocalCache + PathDelim + BaseSceneName + PathDelim + ExtractFilename(SceneContentDirZip)) then
              CopyFile(LnC_Settings.LocalCache + PathDelim + BaseSceneName + PathDelim + ExtractFilename(SceneContentDirZip), SceneContentDirZip)
            else //Else setup the cache directory and place a file in there with the time the download started
              begin
                if (DirectoryExistsUTF8(LnC_Settings.LocalCache)) then //Only do this if the Cache directory exists
                  begin
                    if not DirectoryExistsUTF8(LnC_Settings.LocalCache + PathDelim + BaseSceneName) then CreateDirUTF8(LnC_Settings.LocalCache + PathDelim + BaseSceneName);
                    if not FileExistsUTF8(LnC_Settings.LocalCache + PathDelim + BaseSceneName + PathDelim + 'download.info') then
                      begin
                        //If the download.info file does not exists then create it and begin downloading the content zip file
                        AssignFile(ContetGetTime, LnC_Settings.LocalCache + PathDelim + BaseSceneName + PathDelim + 'download.info');
                        {$I-}
                          Rewrite(ContetGetTime);
                        {$I+}
                        IOR := IOResult;
                        if IOR = 0 then
                          begin
                            //Put the time and the screamer name in there
                            Writeln(ContetGetTime, TimeToStr(Now) + ' - ' + LnC_Settings.ScreamerName);
                            CloseFile(ContetGetTime);
                            LnC_FM.FMJobs_Add('get|' + SceneContentDirZip + '|' + ftpSceneContentDirZip);// Begin the download
                          end;
                      end
                    else //If the download.info file does exist, then check the time the download was set and compare to now
                      begin
                        AssignFile(ContetGetTime, LnC_Settings.LocalCache + PathDelim + BaseSceneName + PathDelim + 'download.info');
                        {$I-}
                          Reset(ContetGetTime);
                        {$I+}
                        IOR := IOResult;
                        if IOR = 0 then
                          begin
                            //Get the time in the file
                            ReadLn (ContetGetTime, s1);
                            CloseFile(ContetGetTime);
                            if s1[2] = ':' then s1 := '0' + s1; //Add zero if first digit is single
                            //Get the current time
                            s2 := TimeToStr(Now);
                            if s2[2] = ':' then s2 := '0' + s2; //Add zero if first digit is single
                            //First check to see if the hours (first two digits) are different
                            if Copy(s1, 1, 2) <> Copy(s2, 1, 2) then
                              begin
                                //Next check to see if it's more than 1 hour off...
                                i := StrToInt(Copy(s1, 1, 2));
                                i2 := StrToInt(Copy(s2, 1, 2));
                                //... if it is, then delete the download.info file so another PC will download the zip file
                                if (i < i2 - 1) or (i > i2) then
                                  begin
                                    if FileExistsUTF8(LnC_Settings.LocalCache + PathDelim + BaseSceneName + PathDelim + 'download.info') then
                                      DeleteFileUTF8(LnC_Settings.LocalCache + PathDelim + BaseSceneName + PathDelim + 'download.info');
                                  end
                                else
                                  begin
                                    //Finally check the minutes and see if the file's minutes are less than the now's minutes...
                                    i := StrToInt(Copy(s1, 4, 2));
                                    i2 := StrToInt(Copy(s2, 4, 2));
                                    //... if it is, then delete the download.info file so another PC will download the zip file
                                    if i < i2 then
                                      if FileExistsUTF8(LnC_Settings.LocalCache + PathDelim + BaseSceneName + PathDelim + 'download.info') then
                                        DeleteFileUTF8(LnC_Settings.LocalCache + PathDelim + BaseSceneName + PathDelim + 'download.info');
                                  end;
                              end;
                          end;
                      end;
                  end
                else //If the local cache dir does not exist, simply download the file
                  LnC_FM.FMJobs_Add('get|' + SceneContentDirZip + '|' + ftpSceneContentDirZip);
              end;
          end;
      end;
    30:
      begin //Wait on render
        Status := 31;
      end;
      31: //Check for output, set to upload and delete
      begin
        //Set output in scene file to local dir
        OutputdirRGB := LnC_Settings.LocalWorkingDir + PathDelim + LnC_GetOutputDir(BaseSceneName, 'rgb');
        OutputdirAlpha := LnC_Settings.LocalWorkingDir + PathDelim + LnC_GetOutputDir(BaseSceneName, 'alpha');
        //Set ftpoutput in scene file to local dir
        ftpOutputdirRGB := LnC_Settings.ftpSubDir + ftpPathDelim + LnC_GetftpOutputDir(BaseSceneName, 'rgb');
        ftpOutputdirAlpha := LnC_Settings.ftpSubDir + ftpPathDelim + LnC_GetftpOutputDir(BaseSceneName, 'alpha');

        //Upload RGB images
        If FindFirst (OutputdirRGB + PathDelim + '*', faAnyFile - faDirectory, OutputFiles) = 0 then
          begin
            Repeat
              With OutputFiles do
                begin
                  LnC_FM.FMJobs_Add('put|' + OutputdirRGB + PathDelim + Name + '|' + ftpOutputdirRGB + ftpPathDelim + Name);
                  LnC_FM.FMJobs_Add('del|' + OutputdirRGB + PathDelim + Name + '|' + 'none');
                end;
            Until FindNext(OutputFiles)<>0;
          end;
        FindClose(OutputFiles);

        //Upload Alpha images
        If FindFirst (OutputdirAlpha + PathDelim + '*', faAnyFile - faDirectory, OutputFiles) = 0 then
          begin
            Repeat
              With OutputFiles do
                begin
                  LnC_FM.FMJobs_Add('put|' + OutputdirAlpha + PathDelim + Name + '|' + ftpOutputdirAlpha + ftpPathDelim + Name);
                  LnC_FM.FMJobs_Add('del|' + OutputdirAlpha + PathDelim + Name + '|' + 'none');
                end;
            Until FindNext(OutputFiles)<>0;
          end;
        FindClose(OutputFiles);
      end;
      40://Finished up the scene
      begin
        //Forward along to upload after this cycle
        Status := 41;
      end;
      41:
      begin //Upload scene output
        NoFilesLeft := True; //Assume there are no files
        //Set output in scene file to local dir
        OutputdirRGB := LnC_Settings.LocalWorkingDir + PathDelim + LnC_GetOutputDir(BaseSceneName, 'rgb');
        OutputdirAlpha := LnC_Settings.LocalWorkingDir + PathDelim + LnC_GetOutputDir(BaseSceneName, 'alpha');
        //Set ftpoutput in scene file to local dir
        ftpOutputdir := LnC_Settings.ftpSubDir + ftpPathDelim + LnC_GetftpOutputDir(BaseSceneName, 'root');
        ftpOutputdirRGB := LnC_Settings.ftpSubDir + ftpPathDelim + LnC_GetftpOutputDir(BaseSceneName, 'rgb');
        ftpOutputdirAlpha := LnC_Settings.ftpSubDir + ftpPathDelim + LnC_GetftpOutputDir(BaseSceneName, 'alpha');

        //Upload RGB images
        If FindFirst (OutputdirRGB + PathDelim + '*', faAnyFile - faDirectory, OutputFiles) = 0 then
          begin
            NoFilesLeft := False;
            Repeat
              With OutputFiles do
                begin
                  LnC_FM.FMJobs_Add('put|' + OutputdirRGB + PathDelim + Name + '|' + ftpOutputdirRGB + ftpPathDelim + Name);
                  LnC_FM.FMJobs_Add('del|' + OutputdirRGB + PathDelim + Name + '|' + '/');
                end;
            Until FindNext(OutputFiles)<>0;
          end;
        FindClose(OutputFiles);

        //Upload Alpha images
        If FindFirst (OutputdirAlpha + PathDelim + '*', faAnyFile - faDirectory, OutputFiles) = 0 then
          begin
            NoFilesLeft := False;
            Repeat
              With OutputFiles do
                begin
                  LnC_FM.FMJobs_Add('put|' + OutputdirAlpha + PathDelim + Name + '|' + ftpOutputdirAlpha + ftpPathDelim + Name);
                  LnC_FM.FMJobs_Add('del|' + OutputdirAlpha + PathDelim + Name + '|' + '/');
                end;
            Until FindNext(OutputFiles)<>0;
          end;
        FindClose(OutputFiles);

        if NoFilesLeft then Status := 42;
      end;
      42:
      begin //Scene is totally finished
        ftpOutputdir := LnC_Settings.ftpSubDir + ftpPathDelim + LnC_GetftpOutputDir(BaseSceneName, 'root');
        LogMsg := 'Finished uploading scene output to ' + ftpOutputdir + ftpPathDelim;
        Status := 43;
      end;
      43:
      begin
        //Do nothing but wait
      end;
      50:
      begin //Reset Data
        ResetScene;
        Status := 0;
      end;
    end;
  end;
end;

function TLWScene.GatherFrames(const SentFrameCount: integer): integer;//Returns number of frames gathered

var
  F: TIniFile;
  i, i2, gatherID: integer;
  FrameCount, strim:integer;
  s, INISceneName, ContentStamp, INIGroupsStr: string;
  FramesToRender: boolean;
  ContentOK,
  ContentOffsiteOK,
  ContentOnsiteOK:boolean;
  KeepCheckingFile: boolean;
  SceneIsOffsite: boolean;
  GroupMatched: boolean;
  INIGroups, LnCGroups:TStringList; // Groups are seperated by spaces do need to be deliminated
  Gi, Gx: integer; // Counters for group checking

begin
  gatherID := 0;
  i  := 0;
  i2 := 1;
  strim:= 0;
  ContentStamp := '';
  if SentFrameCount < 1 then
    FrameCount := 1
  else
    FrameCount := SentFrameCount;

  KeepCheckingFile := True;
  FramesToRender := False;
  ContentOK:=False; //Assume content it bad until found otherwise
  ContentOnsiteOK:= False;//Assume content it bad until found otherwise
  ContentOffsiteOK:= false;//Assume content it bad until found otherwise
  SceneIsOffsite := True; //Assume scene is offsite until found otherwise

  //Get the setting's for files
  ftpSubDir := LnC_Settings.ftpSubDir;
  LocalWorkingDir := LnC_Settings.LocalWorkingDir;

  //Be sure both ini files downloaded, else set them to download
  if (FileExistsUTF8(LocalWorkingDir + PathDelim + OnsiteSceneINI)) and
     (FileExistsUTF8(LocalWorkingDir + PathDelim + OffsiteSceneINI)) then
  begin
    FrameCount := FrameCount - 1; //Needed to make the frames sent properly

    //Check for onsite frames first
    if LnC_Settings.Onsite then
      begin
        F := TIniFile.Create(LocalWorkingDir + PathDelim + OnsiteSceneINI);

        // Check if check for frames returns [none] there is no scene there so stop checking
        while KeepCheckingFile do
          begin
            gatherID := gatherID + 1;
            Str(gatherID, SceneID);
            s  := F.ReadString(SceneID, 'StartFrame', '[none]');
            i  := F.ReadInteger(SceneID, 'StartFrame', 1);
            i2 := F.ReadInteger(SceneID, 'EndFrame', 0);
            INIGroupsStr  := F.ReadString(SceneID, 'Groups', 'default');

            if i <= i2 then
              KeepCheckingFile := False;

            GroupMatched := false; // Assume no match to start
            if not KeepCheckingFile then // Check to see if there are matching groups
              begin
                INIGroups := TStringList.Create;
                INIGroups.DelimitedText := INIGroupsStr;
                LnCGroups := TStringList.Create;
                LnCGroups.DelimitedText := LnC_Settings.Groups;
                for Gi := 0 to LnCGroups.Count - 1 do
                  for Gx := 0 to INIGroups.Count - 1 do
                    if LnCGroups.ValueFromIndex[Gi] = INIGroups.ValueFromIndex[Gx] then
                      GroupMatched := true;
                INIGroups.Free;
                LnCGroups.Free;
                KeepCheckingFile := not GroupMatched; // Keep check would be oposite of a match
              end;

            if s = '[none]' then KeepCheckingFile := False;
          end;

        if s <> '[none]' then FramesToRender := True else FramesToRender := False;

        //If there are frames set scene to onsite for content look-up
        //If no frames to render, then reset the variables for use with Offsite scenes
        if FramesToRender then
          begin
           SceneIsOffsite := false
          end
        else
          begin
            F.Free;
            gatherID := 0;
            i  := 0;
            i2 := 1;
            KeepCheckingFile := true;
            SceneIsOffsite := true;
          end;

      end;//Onsite frame check

    //After checking onsite, then check offsite
    if (LnC_Settings.Offsite) and (not FramesToRender) then
      begin
        F := TIniFile.Create(LocalWorkingDir + PathDelim + OffsiteSceneINI);

        // Check if check for frames returns [none] there is no scene there so stop checking
        while KeepCheckingFile do
          begin
            gatherID := gatherID + 1;
            Str(gatherID, SceneID);
            s  := F.ReadString(SceneID, 'StartFrame', '[none]');
            i  := F.ReadInteger(SceneID, 'StartFrame', 1);
            i2 := F.ReadInteger(SceneID, 'EndFrame', 0);
            INIGroupsStr  := F.ReadString(SceneID, 'Groups', 'default');

            if i <= i2 then
              KeepCheckingFile := False;

            GroupMatched := false; // Assume no match to start
            if not KeepCheckingFile then // Check to see if there are matching groups
              begin
                INIGroups := TStringList.Create;
                INIGroups.DelimitedText := INIGroupsStr;
                LnCGroups := TStringList.Create;
                LnCGroups.DelimitedText := LnC_Settings.Groups;
                for Gi := 0 to LnCGroups.Count - 1 do
                  for Gx := 0 to INIGroups.Count - 1 do
                    if LnCGroups.ValueFromIndex[Gi] = INIGroups.ValueFromIndex[Gx] then
                      GroupMatched := true;
                INIGroups.Free;
                LnCGroups.Free;
                KeepCheckingFile := not GroupMatched; // Keep check would be oposite of a match
              end;

            if s = '[none]' then KeepCheckingFile := False;
          end;

        if s <> '[none]' then FramesToRender := True else FramesToRender := False;

        if FramesToRender then SceneIsOffsite := True else SceneIsOffsite := False;
      end;

    //Define content dir and check if it exists
    if (FramesToRender) and (not SceneIsOffsite) then
      begin
        SceneContentDir := F.ReadString(SceneID, 'SceneContentDir', '');
        if DirectoryExistsUTF8(SceneContentDir) then ContentOnsiteOK := true else ContentOnsiteOK := false;
      end;

    //Check to see if the content directoy has been downloaded
    //if there are frames to render and the scene is offsite
    if (FramesToRender) and (SceneIsOffsite) then
      begin
        SceneContentDir := F.ReadString(SceneID, 'SceneContentDir', '');
        ftpSceneContentDirZip := ftpSubDir + ftpPathDelim + SceneContentDir;
        SceneContentDirZip := LocalWorkingDir + PathDelim + SceneContentDir;
        SceneContentDir := LocalWorkingDir + PathDelim + SceneContentDir;
        strim := Length(SceneContentDir);
        strim := strim - 4;
        SceneContentDir := Copy(SceneContentDir, 1, strim);
        if (DirectoryExistsUTF8(SceneContentDir + ContentDateTimeStamp)) and (BaseSceneName = F.ReadString(SceneID, 'SceneName', '')) then ContentOffsiteOK := true
        else
          begin
            //Go ahead and set base name so the LWD will not be cleared after scene download
            BaseSceneName := F.ReadString(SceneID, 'SceneName', '');
            ContentDateTimeStamp := '';
            LnC_CleanLWD(LnC_Settings.LocalWorkingDir);//New content so clean out LWD
            ContentOffsiteOK := false;
          end;
      end;

    if (FramesToRender) then ContentOK := DirectoryExistsUTF8(SceneContentDir);
    if (FramesToRender) and (SceneIsOffsite) then ContentOK := DirectoryExistsUTF8(SceneContentDir + ContentDateTimeStamp);
// ---- Deal with Scene issues
    if (FramesToRender) and (ContentOK) then
      begin
        INISceneName := F.ReadString(SceneID, 'SceneName', '');
        if INISceneName <> BaseSceneName then
          //Check to see if current scene name is different than the name in the ini file
          begin
            FrameCount := 0; //Then render only one frame
            ContentDateTimeStamp := '';
            LnC_CleanLWD(LnC_Settings.LocalWorkingDir);//And clean LWD
            SceneDled := False;
          end;
        //Now update the Scene Name with the INI Scene Name
        BaseSceneName := INISceneName;
        ftpSceneName := ftpSubDir + ftpPathDelim + INISceneName;
        SceneName := LocalWorkingDir + PathDelim + BaseSceneName;

        //Check to make sure the scene is downloaded
        if SceneDled then
          begin
            //Scene is downloaded and there are frames ready
            i  := F.ReadInteger(SceneID, 'StartFrame', 0);
            i2 := F.ReadInteger(SceneID, 'EndFrame', 0);
            //If local start frame plus FrameCount is less then saved Endframe,
            //then make the final frame sent to the Screamer the Startframe + FrameCount
            //otherwise the final frame will be the EndFrame
            if (i + FrameCount) < i2 then
              i2 := i + FrameCount;

            Str(i, s);
            StartFrame := s;
            Str(i2, s);
            EndFrame := s;

            //Put the end frame plus 1 back into the ini file as the new start frame
            i2 := i2 + 1;
            F.WriteInteger(SceneID, 'StartFrame', i2);

            //Set content directory for offsite scenes
            if SceneIsOffsite then SceneContentDir := SceneContentDir + ContentDateTimeStamp;

            Status := 30;
            LogMsg := 'Frames gathered for ' + BaseSceneName;
            Result := i2 - i;//Return the number of frames actually set to render
          end
        else
          begin
            //Since the Scene file wasn't downloaded, set the Status to 20
            //so that it will be downloaded on the next update
            Status := 20;
            LogMsg := 'Downloading scene: ' + BaseSceneName;
            Result := 0;
          end;
      end;
// ---- Deal with Content issues
    //If there are frames to render in an offsite scene but content is not okay, then dl content
    if (not ContentOffsiteOK) and (FramesToRender) and (SceneIsOffsite) then
      begin
        Status := 21;
        LogMsg := 'Downloading content: ' + ExtractFileName(SceneContentDirZip);
        Result := 0;
        SceneDled := False;
      end;
    //If the scene is onsite, there are frames to render
    //but content directory is not okay, then tell user
    if (not SceneIsOffsite) and (not ContentOK) and (FramesToRender) then
      begin
        LogMsg := 'ERROR: Can not read content directory ' + SceneContentDir;
        Result := 0;
        ResetScene;
        Status := 3;
      end;

    //If there are no frames to render
    if (not FramesToRender) then
      begin
        ResetScene;
        Status := 50;
        Result := 0;
        LogMsg := 'no frames';
       end;
    F.Free;// Finished with INI file
//---------------------------

  if Status = 30 then
     begin
       s := LnC_MakeLastUpdate (LocalWorkingDir + PathDelim, LnC_Settings.ScreamerName);
       //Send the ini files back to the FTP server...
       LnC_FM.FMJobs_Add('put|' + LocalWorkingDir + PathDelim + OnsiteSceneINI + '|' + ftpSubDir + ftpPathDelim + OnsiteSceneINI);
       LnC_FM.FMJobs_Add('put|' + LocalWorkingDir + PathDelim + OffsiteSceneINI + '|' + ftpSubDir + ftpPathDelim + OffsiteSceneINI);
       LnC_FM.FMJobs_Add('put|' + s + '|' + ftpSubDir + ftpPathDelim + ExtractFilename(s));
     end;

  //Delete the ini files for a fresh start on next check
  LnC_FM.FMJobs_Add('del|' + LocalWorkingDir + PathDelim + OnsiteSceneINI + '|' + '/');
  LnC_FM.FMJobs_Add('del|' + LocalWorkingDir + PathDelim + OffsiteSceneINI + '|' + '/');
  end //End of check for ini files
  else //In case the files have not yet been downloaded
    begin
      if Status < 10 then Status := 10;
      if Status >= 50 then Status := 10;
      Result := 0;
      LnC_FM.FMJobs_Add('get|' + LocalWorkingDir + PathDelim + OnsiteSceneINI + '|' + ftpSubDir + ftpPathDelim + OnsiteSceneINI);
      LnC_FM.FMJobs_Add('get|' + LocalWorkingDir + PathDelim + OffsiteSceneINI + '|' + ftpSubDir + ftpPathDelim + OffsiteSceneINI);
    end;
end;

procedure TLWScene.Finish;
begin
  Status := 40;
end;

procedure TLWScene.ResetScene;
begin
  SceneID    := '';
  SceneContentDir := '';
  BaseSceneName := '';
  SceneName  := '';
  StartFrame := '';
  EndFrame   := '';
  Status     := 0;
  SceneDled := False;
//  LogMsg     := 'Scene Reset';
end;

procedure TLWScene.Uploadlog(logfile: string);

begin
  LnC_FM.FMJobs_Add('put|' + LocalWorkingDir + PathDelim + logfile + '|' + ftpSubDir + ftpPathDelim + logfile);
end;

function TLWScene.IsOperational: boolean;
begin
  if Status > 9 then
    Result := True
  else
    Result := False;
end;

function TLWScene.IsTryingToGetSceneFiles: boolean;

begin
  if (Status > 19) and (Status < 30) then
    Result := True
  else
    Result := False;
end;

function TLWScene.HasGatheredScene: boolean;
begin
  if (Status > 29) and (Status < 40) then
    Result := True
  else
    Result := False;
end;

function TLWScene.ReadyToGather: boolean;
begin
  if Status = 10 then
    Result := True
  else
    Result := False;
end;

function TLWScene.IsFinished: boolean;
begin
  if Status = 43 then
    Result := True
  else
    Result := False;
end;

end.

