{
  Copyright 2011-2012 Joe Justice (joe@joejustice.org)
 
  This file is part of LightNet: Cloud.

  LightNet: Cloud is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  LightNet: Cloud is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with LightNet: Cloud, see files COPYING.txt and COPYING.LESSER.txt 
  If not, see <http://www.gnu.org/licenses/>.

----------------------------------------------------------------------------

Performs all communication with LightWave ScreamerNet.

}
unit lwscreamer_unit;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Process, FileUtil;

type
  TLWScreamer  = class(TObject)
    public
      //Message for log
      LogMsg: string;

      //Communications with LnC
      function IsOperational: boolean;
      function IsRendering: boolean;
      function IsFirstWait: boolean;
      function IsFinishedRendering: boolean;
      function IsWaiting: boolean;

      //Commands from LnC
      procedure Launch (const lwsnexe, configdir, commanddir:string; lwSilent: boolean);
      procedure Kill;
      procedure Stop;
      procedure StatusUpdate;
      procedure Render(const ContentDir: string; SceneFile: string; StartFrame: string; EndFrame: string);

      constructor Create;
      destructor Destroy; override;
    private
      Status: integer;
    { Status values:
      Using units of 10 so as to have room for various sub-statuses
      0 - Not Operational
      10 - Initialized
      20 - Ready for the first time
      21 - Ready to recieve a scene
      22 - First launch wait
      30 - Changing Content Directory
      31 - Just changed Content Directoy
      40 - Loading command sent
      41 - Scene is loaded
      42 - New frames have been sent, but scene is already loaded
      50 - Rendering command sent
      51 - Finished rendering
      60 - Shutting Down
      }

      LWSN_EXE: TProcess;

      ShutdownScream: Boolean;

      AtLaunchCommandDir: string;

      ActiveScene, ActiveContentDir,
      ActiveFirstFrame, ActiveEndFrame: string;

      procedure ResetScreamer;
      function WriteJob (Job: string; JobDir: string): boolean;
      function ReadAck (AckDir: string): string;

  end;

implementation

constructor TLWScreamer.Create;
  begin
    Status := 0;
    LWSN_EXE := TProcess.Create(nil);
    LogMsg := '';
    ActiveScene := '';
    ActiveContentDir := '';
    ActiveFirstFrame := '';
    ActiveEndFrame := '';
    ShutdownScream := False;
  end;

destructor TLWScreamer.Destroy;
  begin
    LWSN_EXE.Free;
  end;

procedure TLWScreamer.ResetScreamer;
  begin
    Status := 0;
    ActiveScene := '';
    ActiveContentDir := '';
    ActiveFirstFrame := '';
    ActiveEndFrame := '';
    ShutdownScream := False;
  end;

procedure TLWScreamer.Kill;
  var
    i: integer;

  begin
    i := 0;
    if LWSN_EXE.Running then LWSN_EXE.Terminate(i);
    Status := 0;
    LogMsg := 'Terminated';
    ActiveScene := '';
    ActiveContentDir := '';
    ActiveFirstFrame := '';
    ActiveEndFrame := '';
  end;

procedure TLWScreamer.StatusUpdate;
  var
    AckReply: string;

  begin
    //Check to be sure Status is > 9 before doing anything else
    if Status > 9 then
      begin
        AckReply := ReadAck(AtLaunchCommandDir);

        case Status of
          10: begin //Recieve Initializing acknowledgment, send wait command
                if (AckReply = 'Initializing') or (AckReply = 'Ready') then
                  begin
                    WriteJob ('wait', AtLaunchCommandDir);
                    LogMsg := 'ScreamerNet initialized';
                    Status := 20;
                  end;
              end;
          20: //Recieve Ready acknowledgment for first time, update status to standby
            begin
              if AckReply = 'Ready' then
                begin
                  Status := 22;
                  //LogMsg := 'Screamer Ready';
                end;
            end;
          21: //Waiting to render
            begin
              //Do nothing at this time but wait.
              if ShutdownScream then
                begin
                  WriteJob ('quit', AtLaunchCommandDir);
                  ShutdownScream := False;
                  Status := 60;
                end;
            end;
          22:
            begin
              //First launch wait
              if ShutdownScream then
                begin
                  WriteJob ('quit', AtLaunchCommandDir);
                  ShutdownScream := False;
                  Status := 60;
                end
              else Status := 21;
            end;
          30:
            begin
              if AckReply = 'Setting content directory' then
                begin
                  WriteJob ('wait', AtLaunchCommandDir);
                  Status := 31;
//                  LogMsg := 'ScreamerNet set content to ' + ActiveContentDir;
                end;
            end;
          31: //Recieve ready acknowledgment, send load command
            begin
              if AckReply = 'Ready' then
                begin
                  WriteJob ('load ' + ActiveScene, AtLaunchCommandDir);
                  //LogMsg := 'Ready to load ' + ActiveScene;
                  Status := 40;
                end;
            end;
          40: //Recieve loading acknowledgment, send load command
            begin
              if AckReply = 'Loading' then
                begin
                  WriteJob ('wait', AtLaunchCommandDir);
//                  LogMsg := 'ScreamerNet loaded ' + ActiveScene;
                  Status := 41;
                end;
            end;
          41: //Recieve ready acknowledgment, send render command
            begin
              if AckReply = 'Ready' then
                begin
                  WriteJob ('render ' + ActiveFirstFrame + ' ' + ActiveEndFrame + ' 1', AtLaunchCommandDir);
                  //LogMsg := ActiveScene + ' loaded';
                  Status := 50;
                end;
             end;
          42: //If same scene with different frames, this will be the direct render bypassing content and load commands
            begin
              if AckReply = 'Ready' then
                begin
                  WriteJob ('render ' + ActiveFirstFrame + ' ' + ActiveEndFrame + ' 1', AtLaunchCommandDir);
//                  LogMsg := 'ScreamerNet rendering more frames from same scene';
                  Status := 50;
                end;
             end;
          50:  //Recieve Rendering frame... acknowledgment, send wait command
            begin
              if AckReply = ('Rendering frame ' + ActiveFirstFrame) then
                begin
                  WriteJob ('wait', AtLaunchCommandDir);
                  LogMsg := 'ScreamerNet rendering frame(s) ' + ActiveFirstFrame + ' - ' + ActiveEndFrame;
                  Status := 51;
                end;
            end;
          51:  //Recieve ready acknowledgment, reset status to waiting
            begin
              if AckReply = 'Ready' then
                begin
//                  LogMsg := 'Frame(s) ' + ActiveFirstFrame + '-' + ActiveEndFrame + ' complete by ScreamerNet';
                  Status := 52;
                end;
            end;
          52: //Is Finished rendering and has responded
            begin
              if ShutdownScream then
                begin
                  WriteJob ('quit', AtLaunchCommandDir);
                  ShutdownScream := False;
                  Status := 60;
                end
              else Status := 21;
            end;
          60:
            begin
              //Pause for one cycle in order for the scene to clean-up
              Status := 61;
            end;
          61:
            begin
              if AckReply = 'Offline' then
                begin
                  WriteJob ('wait', AtLaunchCommandDir);
                  LogMsg := 'Offline';
                  ResetScreamer;
                end;
            end;

        end; // End Case
     //In case user has shut down LW Screamernet manually, set status to 0
    if not LWSN_EXE.Running then
      begin
        Status := 0;
        if AckReply = 'Offline' then LogMsg := 'Offline'
          else LogMsg := 'ERROR: crashed';
        ResetScreamer;
      end;

      end;
  end;

procedure TLWScreamer.Stop;
  begin
    ShutdownScream := True;
    LogMsg := 'ScreamerNet shutting down when finished...';
  end;

procedure TLWScreamer.Launch (const lwsnexe, configdir, commanddir: string; lwSilent: boolean);
  var
    SettingsError:boolean;
  begin
    SettingsError := false;

    if (not DirectoryExistsUTF8(configdir)) or
       (not FileExistsUTF8(lwsnexe)) or
       (not DirectoryExistsUTF8(commanddir))
       then SettingsError := true;

    if (not LWSN_EXE.Running) and (not SettingsError) then
      begin
        AtLaunchCommandDir := commanddir;

        //init job file
        WriteJob ('init', AtLaunchCommandDir);

        LWSN_EXE.CommandLine := lwsnexe + ' -2 -c"'
                                + configdir + '" "'
                                + commanddir + PathDelim+'job" ' + ' "'
                                + commanddir + PathDelim+'ack"';

        if lwSilent then LWSN_EXE.Priority := ppIdle else LWSN_EXE.Priority := ppNormal;
        if lwSilent then LWSN_EXE.ShowWindow := swoShowMinimized else LWSN_EXE.ShowWindow := swoNone;

        LWSN_EXE.Execute;
        Status := 10; // Initializing
      end;
      if SettingsError then
        begin
          LogMsg := 'ERROR: settings';
          Status := 0; // Initializing
        end;
  end;

function TLWScreamer.WriteJob (Job: string; JobDir: string): boolean;
  var
    JobFile: TextFile;

  begin
    try
      AssignFile(JobFile, JobDir+PathDelim+'job');
      Rewrite(JobFile);
      Writeln(JobFile,Job);
    finally
      CloseFile(JobFile);
    end;

    Result := true; //*** Need to create error catcher ***
  end;

function TLWScreamer.ReadAck (AckDir: string): string;
  var
    AckFile: TextFile;
    IOR: Integer;

  begin
    try
      AssignFile(AckFile, AckDir+PathDelim+'ack');
      {$I-}
      Reset(AckFile);
      {$I+}
      IOR := IOResult;
      if IOR <> 0 then
        begin
          Result := '';
        end
      else
        begin
          ReadLn (AckFile, Result);
          CloseFile(AckFile);
        end;
    finally
    end;
  end;

procedure TLWScreamer.Render(const ContentDir: string; SceneFile: string; StartFrame: string; EndFrame: string);
  begin
  { **************************************************************
      Writing the content job and setting Status to 30 sets the
      Content -> Load -> Render -> Wait
      process into motion.

      Once it begins. There is not stopping it! (cue daramatic music)

    **************************************************************
  }
    if SceneFile = ActiveScene then //Check with this new command, see if the scene is already loaded...
      begin
        Status := 42;  //...if so, set Status to bypase loading
        ActiveScene := SceneFile;
        ActiveContentDir := ContentDir;
        ActiveFirstFrame := StartFrame;
        ActiveEndFrame := EndFrame;
      end
    else //If this is a new scene...
      begin
        WriteJob ('content ' + ContentDir, AtLaunchCommandDir);
        Status := 30;
        ActiveScene := SceneFile;
        ActiveContentDir := ContentDir;
        ActiveFirstFrame := StartFrame;
        ActiveEndFrame := EndFrame;
      end;
  end;

function TLWScreamer.IsOperational: boolean;
  begin
    if Status > 9 then Result := true else Result := false;
  end;

function TLWScreamer.IsRendering: boolean;
  begin
    if (Status > 30) and (Status < 70) then Result := true else Result := false;
  end;

function TLWScreamer.IsFinishedRendering: boolean;
  begin
    if Status = 52 then Result := true else Result := false;
  end;

function TLWScreamer.IsWaiting: boolean;
  begin
    if Status = 21 then Result := true else Result := false;
  end;

function TLWScreamer.IsFirstWait: boolean;
  begin
    if Status = 22 then Result := true else Result := false;
  end;

end.

